"""
An IP selector implementation returning client IP from REMOTE_ADDR.
"""

from .interface import IPSelectorInterface
from ipinfo_django.helpers import REMOTE_ADDR


class ClientIPSelector(IPSelectorInterface):
    """Use client IP address from REMOTE_ADDR."""

    def get_ip(self, request):
        return request.META.get(REMOTE_ADDR)



"""
Abstract interface for selecting IP from incoming request.
"""

import abc


class IPSelectorInterface(metaclass=abc.ABCMeta):
    """Interface for selecting IP."""

    @abc.abstractmethod
    def get_ip(self, request):
        pass



import logging
import traceback

import ipinfo
from django.conf import settings

from ipinfo_django.helpers import is_bot
from ipinfo_django.ip_selector.default import DefaultIPSelector

LOGGER = logging.getLogger(__name__)


class IPinfoMiddleware:
    def __init__(self, get_response=None):
        """
        Initializes class while gettings user settings and creating the cache.
        """
        self.get_response = get_response
        self.filter = getattr(settings, "IPINFO_FILTER", self.is_bot)

        ipinfo_token = getattr(settings, "IPINFO_TOKEN", None)
        ipinfo_settings = getattr(settings, "IPINFO_SETTINGS", {})
        self.ip_selector = getattr(settings, "IPINFO_IP_SELECTOR", DefaultIPSelector())
        self.ipinfo = ipinfo.getHandler(ipinfo_token, **ipinfo_settings)

    def __call__(self, request):
        """Middleware hook that acts on and modifies request object."""
        try:
            if self.filter and self.filter(request):
                request.ipinfo = None
            else:
                request.ipinfo = self.ipinfo.getDetails(
                    self.ip_selector.get_ip(request)
                )
        except Exception:
            request.ipinfo = None
            LOGGER.error(traceback.format_exc())

        response = self.get_response(request)
        return response

    def is_bot(self, request):
        return is_bot(request)


class IPinfoAsyncMiddleware:
    sync_capable = False
    async_capable = True

    def __init__(self, get_response):
        """Initialize class, get settings, and create the cache."""
        self.get_response = get_response

        self.filter = getattr(settings, "IPINFO_FILTER", self.is_bot)

        ipinfo_token = getattr(settings, "IPINFO_TOKEN", None)
        ipinfo_settings = getattr(settings, "IPINFO_SETTINGS", {})
        self.ip_selector = getattr(settings, "IPINFO_IP_SELECTOR", DefaultIPSelector())
        self.ipinfo = ipinfo.getHandlerAsync(ipinfo_token, **ipinfo_settings)

    def __call__(self, request):
        return self.__acall__(request)

    async def __acall__(self, request):
        """Middleware hook that acts on and modifies request object."""
        try:
            if self.filter and self.filter(request):
                request.ipinfo = None
            else:
                request.ipinfo = await self.ipinfo.getDetails(
                    self.ip_selector.get_ip(request)
                )
        except Exception:
            request.ipinfo = None
            LOGGER.error(traceback.format_exc())

        response = await self.get_response(request)
        return response

    def is_bot(self, request):
        return is_bot(request)


class IPinfoLiteMiddleware(IPinfoMiddleware):
    def __init__(self, get_response):
        super().__init__(get_response=get_response)

        ipinfo_token = getattr(settings, "IPINFO_TOKEN", None)
        ipinfo_settings = getattr(settings, "IPINFO_SETTINGS", {})
        self.ipinfo = ipinfo.getHandlerLite(ipinfo_token, **ipinfo_settings)


class IPinfoAsyncLiteMiddleware(IPinfoAsyncMiddleware):
    sync_capable = False
    async_capable = True

    def __init__(self, get_response):
        super().__init__(get_response=get_response)
        ipinfo_token = getattr(settings, "IPINFO_TOKEN", None)
        ipinfo_settings = getattr(settings, "IPINFO_SETTINGS", {})
        self.ipinfo = ipinfo.getHandlerAsyncLite(ipinfo_token, **ipinfo_settings)



HTTP_X_FORWARDED_FOR = "HTTP_X_FORWARDED_FOR"
REMOTE_ADDR = "REMOTE_ADDR"


def is_bot(request):
    """Whether or not the request user-agent self-identifies as a bot"""
    lowercase_user_agent = request.META.get("HTTP_USER_AGENT", "").lower()
    return "bot" in lowercase_user_agent or "spider" in lowercase_user_agent



from http import HTTPStatus
from unittest import mock

from ipinfo.details import Details


def test_middleware_appends_ip_info(client, ipinfo_lite_middleware):
    with mock.patch("ipinfo.HandlerLite.getDetails") as mocked_getDetails:
        mocked_getDetails.return_value = Details({"ip": "127.0.0.1"})
        res = client.get("/test_view/")
        assert res.status_code == HTTPStatus.OK
        assert b"For testing: 127.0.0.1" in res.content


def test_middleware_filters(client, ipinfo_lite_middleware):
    res = client.get("/test_view/", HTTP_USER_AGENT="some bot")
    assert res.status_code == HTTPStatus.OK
    assert b"Request filtered." in res.content


def test_middleware_behind_proxy(client, ipinfo_lite_middleware):
    with mock.patch("ipinfo.HandlerLite.getDetails") as mocked_getDetails:
        mocked_getDetails.return_value = Details({"ip": "93.44.186.197"})
        res = client.get("/test_view/", HTTP_X_FORWARDED_FOR="93.44.186.197")

        mocked_getDetails.assert_called_once_with("93.44.186.197")
        assert res.status_code == HTTPStatus.OK
        assert b"For testing: 93.44.186.197" in res.content


def test_middleware_not_behind_proxy(client, ipinfo_lite_middleware):
    with mock.patch("ipinfo.HandlerLite.getDetails") as mocked_getDetails:
        mocked_getDetails.return_value = Details({"ip": "127.0.0.1"})
        res = client.get("/test_view/")

        mocked_getDetails.assert_called_once_with("127.0.0.1")
        assert res.status_code == HTTPStatus.OK
        assert b"For testing: 127.0.0.1" in res.content



"""
A default iphandler implementation that returns IP to query for depending on whether we are behind a reverse proxy or not.
"""

from .interface import IPSelectorInterface
from ipinfo_django.helpers import HTTP_X_FORWARDED_FOR, REMOTE_ADDR


class DefaultIPSelector(IPSelectorInterface):
    """Default IP selector which uses IP depending on whether we are behind a reverse proxy or not."""

    def get_ip(self, request):
        """Determine what IP to query for depending on whether we are behind a reverse proxy or not."""
        x_forwarded_for = request.META.get(HTTP_X_FORWARDED_FOR)
        if x_forwarded_for:
            return x_forwarded_for.split(",")[0]
        return request.META.get(REMOTE_ADDR)
